/*

AAC bitrate grapher

produces bitrate histogram of MPEG-4 AAC files for gnuplot
Copyright 2007 tmkk
Inspired by vinfo by Aoyumi

*/

#include <stdio.h>
#include <unistd.h>
#include <AudioToolbox/AudioToolbox.h>

#ifdef _BIG_ENDIAN
#define SWAP32(n) (n)
#define SWAP16(n) (n)
#else
#define SWAP32(n) (((n>>24)&0xff) | ((n>>8)&0xff00) | ((n<<8)&0xff0000) | ((n<<24)&0xff000000))
#define SWAP16(n) (((n>>8)&0xff) | ((n<<8)&0xff00))
#endif

int main(int argc, char *argv[])
{
	ExtAudioFileRef file;
	FSRef inputFSRef;
	FILE *fp,*fpw1,*fpw2;
	int i,tmp,time,frameNum,sample,resolution;
	double frameSize;
	char atom[4],filename[1024];
	char *ptr;
	
	if(argc<2 || argc > 3) {
		fprintf(stderr,"usage: %s aacfile.m4a [resolution]\n",argv[0]);
		return 0;
	}
	
	FSPathMakeRef((UInt8 *)argv[1],&inputFSRef,NULL);
	if(ExtAudioFileOpen(&inputFSRef, &file) != noErr) {
		fprintf(stderr,"can't open file\n");
		return 0;
	}
	AudioStreamBasicDescription inputFormat;
	UInt32 size = sizeof(inputFormat);
	if(ExtAudioFileGetProperty(file, kExtAudioFileProperty_FileDataFormat, &size, &inputFormat) != noErr) {
		ExtAudioFileDispose(file);
		fprintf(stderr,"broken file\n");
		return 0;
	}
	if(inputFormat.mFormatID != kAudioFormatMPEG4AAC) {
		ExtAudioFileDispose(file);
		fprintf(stderr,"not AAC file\n");
		return 0;
	}
	SInt64 frames;
	size = sizeof(frames);
	if(ExtAudioFileGetProperty(file, kExtAudioFileProperty_FileLengthFrames, &size, &frames) != noErr) {
		ExtAudioFileDispose(file);
		fprintf(stderr,"broken file\n");
		return 0;
	}
	double sec = (double)frames / inputFormat.mSampleRate;
	ExtAudioFileDispose(file);
	
	if(argc == 3 && atoi(argv[2]) >= 1024) resolution = atoi(argv[2]);
	else resolution = inputFormat.mSampleRate;
	
	for(ptr=argv[1]+strlen(argv[1]);ptr!=argv[1];ptr--) {
		if(*ptr == '/') {
			ptr++;
			break;
		}
	}
	
	fp = fopen(argv[1], "rb");
	sprintf(filename,"%s_bitrate.plt",ptr);
	fpw1 = fopen(filename,"wb");
	sprintf(filename,"%s_bitrate.txt",ptr);
	fpw2 = fopen(filename,"wb");
	if(!fpw1 || !fpw2) {
		fprintf(stderr,"can't write bitrate file\n");
		goto end;
	}
	
	getcwd(filename,1024);
	sprintf(filename,"%s/%s_bitrate.txt",filename,ptr);
	
	while(1) { //skip until mdat;
		if(fread(&tmp,4,1,fp) < 1) goto end;
		if(fread(atom,1,4,fp) < 4) goto end;
		tmp = SWAP32(tmp);
		if(!memcmp(atom,"mdat",4)) break;
		if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end;
	}
	int bitrate = round((tmp-8)/sec*8) ;
	//fprintf(stderr,"%f,%d\n",sec,bitrate);
	rewind(fp);
	
	fprintf(fpw1,"set title '%s'\n",argv[1]);
	fprintf(fpw1,"set grid\n");
	if(resolution == (int)inputFormat.mSampleRate) fprintf(fpw1,"set xlabel \"time\"\n");
	else fprintf(fpw1,"set xlabel \"time [sec]\"\n");
	fprintf(fpw1,"set ylabel \"bitrate [bps]\"\n");
	fprintf(fpw1,"average = %d\n",bitrate);
	if(resolution == (int)inputFormat.mSampleRate) {
		fprintf(fpw1,"set xdata time\n");
		fprintf(fpw1,"set timefmt \"%%M:%%S\"\n");
	}
	fprintf(fpw1,"plot '%s' using 1:2 w l title 'bitrate', average\n",filename);
	fclose(fpw1);
	
	while(1) { //skip until moov;
		if(fread(&tmp,4,1,fp) < 1) goto end;
		if(fread(atom,1,4,fp) < 4) goto end;
		tmp = SWAP32(tmp);
		if(!memcmp(atom,"moov",4)) break;
		if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end;
	}
	
	while(1) { //skip until trak;
		if(fread(&tmp,4,1,fp) < 1) goto end;
		if(fread(atom,1,4,fp) < 4) goto end;
		tmp = SWAP32(tmp);
		if(!memcmp(atom,"trak",4)) break;
		if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end;
	}
	
	while(1) { //skip until mdia;
		if(fread(&tmp,4,1,fp) < 1) goto end;
		if(fread(atom,1,4,fp) < 4) goto end;
		tmp = SWAP32(tmp);
		if(!memcmp(atom,"mdia",4)) break;
		if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end;
	}
	
	while(1) { //skip until minf;
		if(fread(&tmp,4,1,fp) < 1) goto end;
		if(fread(atom,1,4,fp) < 4) goto end;
		tmp = SWAP32(tmp);
		if(!memcmp(atom,"minf",4)) break;
		if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end;
	}
	
	while(1) { //skip until stbl;
		if(fread(&tmp,4,1,fp) < 1) goto end;
		if(fread(atom,1,4,fp) < 4) goto end;
		tmp = SWAP32(tmp);
		if(!memcmp(atom,"stbl",4)) break;
		if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end;
	}
	
	while(1) { //skip until stsz;
		if(fread(&tmp,4,1,fp) < 1) goto end;
		if(fread(atom,1,4,fp) < 4) goto end;
		tmp = SWAP32(tmp);
		if(!memcmp(atom,"stsz",4)) break;
		if(fseek(fp,tmp-8,SEEK_CUR) != 0) goto end;
	}
	if(fseek(fp,8,SEEK_CUR) != 0) goto end;
	
	if(fread(&tmp,4,1,fp) < 1) goto end;
	frameNum = SWAP32(tmp);
	
	if(resolution == (int)inputFormat.mSampleRate) {
		for(i=0,sample=0,frameSize=0,time=1;i<frameNum;i++) {
			if(fread(&tmp,4,1,fp) < 1) goto end;
			tmp = SWAP32(tmp);
			if(sample + 1024 >= (int)inputFormat.mSampleRate) {
				frameSize += (inputFormat.mSampleRate - sample)/1024.0 * tmp;
				fprintf(fpw2,"%02d:%02d\t%f\n",time/60, time-60*(time/60), frameSize *8);
				sample = 1024 - ((int)inputFormat.mSampleRate - sample);
				frameSize = sample/1024.0 * tmp;
				time++;
				continue;
			}
			frameSize += tmp;
			sample += 1024;
		}
	}
	else {
		for(i=0,sample=0,frameSize=0,time=1;i<frameNum;i++) {
			if(fread(&tmp,4,1,fp) < 1) goto end;
			tmp = SWAP32(tmp);
			if(sample + 1024 >= resolution) {
				frameSize += (resolution - sample)/1024.0 * tmp;
				fprintf(fpw2,"%f\t%f\n",resolution*time/inputFormat.mSampleRate, frameSize * (inputFormat.mSampleRate/resolution) * 8);
				sample = 1024 - (resolution - sample);
				frameSize = sample/1024.0 * tmp;
				time++;
				continue;
			}
			frameSize += tmp;
			sample += 1024;
		}
	}
  end:
	
	fclose(fpw1);
	fclose(fpw2);
	fclose(fp);
	return 0;
}